unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, OleCtrls, SHDocVw, StdCtrls, ActiveX, JPEG, ComObj, MSHTML_TLB,
  ExtCtrls, StrUtils, OpenGL;
  
const
  OpenGLTextureSize = 512;
  WikipediaWidth = 160;
  ScrollbarWidth = 20;
  HeaderHeight = 50 + 128;
  FooterHeight = 156;

  { All three of these are used to try find the biography of current artist. }
  WikiRoot = 'en.wikipedia.org';
  WikiPath = '/wiki/Special:Search?search=';
  WikiParams ='&go=Go';

type
  PRGBArray = ^TRGBArray;
  TRGBArray = array [0..0] of TRGBTriple;

  TfrmMain = class(TForm)
    wbBrowser : TWebBrowser;
    pOGL : TPanel;
    tOGL : TTimer;
    eArtist: TEdit;
    Label1: TLabel;
    Label2: TLabel;
    procedure FormShow(Sender : TObject);
    procedure FormCreate(Sender : TObject);
    procedure tOGLTimer(Sender : TObject);
    procedure wbBrowserDocumentComplete (Sender : TObject; const pDisp : IDispatch; var URL : OleVariant);
    procedure FormKeyDown (Sender : TObject; var Key : Word; Shift : TShiftState);
    procedure eArtistKeyDown (Sender : TObject; var Key : Word; Shift : TShiftState);
  private
    procedure SaveWikipediaBiography (Browser : iWebBrowser2; W, H : Integer);
    procedure Draw;
    procedure CopySubBitmapToTexture;
  public
  end;

var
  frmMain : TfrmMain;
  BiographyTexture : GLUInt;
  SourceBitmap, GeneratedBitmap : TBitmap;
  GeneratedByteArray : array of Byte;
  YOffset : Integer;
  P : TPoint;


implementation

{$R *.dfm}

procedure glBindTexture (Target : GLEnum; Texture : GLUInt); stdcall; external OpenGL32;
procedure glGenTextures (N : GLSizeI; var Textures : GLUInt); stdcall; external OpenGL32;


function GetBrowserElement ( Browser : TWebBrowser;
                             out Element : IHTMLElement2
                           ) : HResult;
var
  pDoc : IHTMLDocument2;
  pElemParent, pElement : IHTMLElement;
  pElement2 : IHTMLElement2;
begin
  if (Assigned (Browser.Document)) then
  begin
    Result := Browser.Document.QueryInterface (IHTMLDocument2,pDoc);
    if (Succeeded (Result)) then
    begin
      try
        if (Assigned (pDoc.Body)) then
        begin
          Result := pDoc.Body.QueryInterface (IHTMLElement,pElement);
          if (Succeeded (Result)) then
          begin
            try
              pElemParent := pElement.parentElement;
              if (Assigned (pElemParent)) then
              begin
                try
                  if (Succeeded (pElemParent.QueryInterface (IHTMLElement2,pElement2))) then
                  begin
                    if (pElement2.clientWidth = 0) or (pElement2.clientHeight = 0) then
                    begin
                      Result := pElement.QueryInterface (IHTMLElement2,Element);
                    end else
                    begin
                      Element := pElement2;
                    end;
                  end else
                  begin
                    Result := pElement.QueryInterface (IHTMLElement2,Element);
                  end;
                finally
                  pElemParent := nil;
                end;
              end else
              begin
                Result := E_NOINTERFACE;
              end;
            finally
              pElement := nil;
            end;
          end;
        end else
        begin
          Result := E_NOINTERFACE;
        end;
      finally
         pDoc := nil;
      end;
    end;
  end else
  begin
    Result := E_NOINTERFACE;
  end;
end;


function GetDocumentSize ( Browser : TWebBrowser
                         ) : TPoint;
var
  pElement : IHTMLElement2;
begin
  if Succeeded (GetBrowserElement (Browser,pElement)) then
  begin
    try
      Result := Point (pElement.ScrollWidth,pElement.ScrollHeight);
    finally
      pElement := nil;
    end;
  end;
end;


procedure SetupPixelFormat ( DC : HDC
                           );
const
   PFD : TPIXELFORMATDESCRIPTOR = (
         nSize : SizeOf (TPIXELFORMATDESCRIPTOR); nVersion : 1;
         dwFlags : PFD_SUPPORT_OPENGL or PFD_DRAW_TO_WINDOW or PFD_DOUBLEBUFFER;
         iPixelType : PFD_TYPE_RGBA;
         cColorBits : 24; cRedBits : 0; cRedShift : 0; cGreenBits : 0; cGreenShift : 0;
         cBlueBits : 0; cBlueShift : 0; cAlphaBits : 0; cAlphaShift : 0; cAccumBits : 0;
         cAccumRedBits : 0; cAccumGreenBits : 0; cAccumBlueBits : 0; cAccumAlphaBits : 0;
         cDepthBits : 16; cStencilBits : 0; cAuxBuffers : 0; iLayerType : PFD_MAIN_PLANE;
         bReserved : 0; dwLayerMask : 0; dwVisibleMask : 0; dwDamageMask : 0);
var
  PixelFormat : Integer;
begin
   PixelFormat := ChoosePixelFormat (DC,@PFD);
   if (PixelFormat = 0) then
   begin
     Exit;
   end;

   if (SetPixelFormat (DC,PixelFormat,@PFD) <> True) then
   begin
     Exit;
   end;
end;


procedure TfrmMain.FormCreate ( Sender : TObject
                              );
var
  DC : HDC;
  RC : HGLRC;
begin
  DC := GetDC (pOGL.Handle);
  SetupPixelFormat (DC);
  RC := wglCreateContext (DC);
  wglMakeCurrent (DC,RC);

  glClearColor (0,0,0.5,1);
  glDepthFunc (GL_LEQUAL);
  glEnable (GL_DEPTH_TEST);

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity;
  gluPerspective (45,pOGL.Width / pOGL.Height,1,100);
  glMatrixMode (GL_MODELVIEW);

  { This will be a constant size... set it once here. }
  SetLength (GeneratedByteArray,((OpenGLTextureSize * OpenGLTextureSize) * 3) - 1);

  ActiveControl := eArtist;
end;


procedure TfrmMain.FormShow ( Sender : TObject
                            );
var
  URL : String;
begin
  URL := WikiRoot + WikiPath + ANSIReplaceText (eArtist.Text,' ','+') + WikiParams;
  wbBrowser.Navigate (URL);
end;


procedure TfrmMain.SaveWikipediaBiography ( Browser : iWebBrowser2;
                                            W, H : Integer
                                          );
var
  SourceDrawRect : TRect;
  ViewObject : IViewObject;
begin
  SourceBitmap := TBitmap.Create;
  SourceDrawRect := Rect (0,0,W,H);
  SourceBitmap.Width := W;
  SourceBitmap.Height := H;

  ViewObject := Browser as IViewObject;
  ViewObject.Draw (DVASPECT_CONTENT,1,nil,nil,0,SourceBitmap.Canvas.Handle,@SourceDrawRect,nil,nil,0);

  { Crop out the left-hand side Wikipedia banner... }
  SourceBitmap.Canvas.CopyRect (Rect (0,0,W - (WikipediaWidth + ScrollbarWidth),H - (HeaderHeight + FooterHeight)),SourceBitmap.Canvas,Rect (WikipediaWidth,HeaderHeight,W - ScrollbarWidth,H - FooterHeight));
  SourceBitmap.Width := W - (WikipediaWidth + ScrollBarWidth);
  SourceBitmap.Height := H - (HeaderHeight + FooterHeight);

  glPixelStoreI (GL_UNPACK_ALIGNMENT,1);
  glGenTextures (1,BiographyTexture);
  glBindTexture (GL_TEXTURE_2D,BiographyTexture);
  glTexParameterI (GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameterI (GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

  CopySubBitmapToTexture;
end;


procedure TfrmMain.Draw;
begin
  glClear (GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glLoadIdentity;
  glTranslatef (0,0,-2.6);
  glRotatef (Sin (GetTickCount / 500) * 6,0,1,0);
  glRotatef (Sin (GetTickCount / 600) * 3,1,0,0);

  glBindTexture (GL_TEXTURE_2D,BiographyTexture);
  glBegin (GL_QUADS);
    glTexCoord2f (0,0);
    glVertex2f (-1,-1);
    glTexCoord2f (1,0);
    glVertex2f (1,-1);
    glTexCoord2f (1,1);
    glVertex2f (1,1);
    glTexCoord2f (0,1);
    glVertex2f (-1,1);
  glEnd;

  SwapBuffers (wglGetCurrentDC);
end;


procedure TfrmMain.tOGLTimer ( Sender : TObject
                             );
begin
  Draw;
end;


procedure TfrmMain.wbBrowserDocumentComplete ( Sender : TObject;
                                               const pDisp : IDispatch;
                                               var URL : OleVariant
                                             );
var
  IDoc1 : IHTMLDocument2;
  Web : IWebBrowser2;
begin
  with wbBrowser do
  begin
    Document.QueryInterface (IHTMLDocument2,iDoc1);
    Web := ControlInterface;
    TControl (wbBrowser).Visible := Boolean (0);
    P := GetDocumentSize (wbBrowser);
    Width := OpenGLTextureSize + WikipediaWidth + ScrollbarWidth;
    Height := P.Y;
    SourceBitmap.Free;
    SaveWikipediaBiography (Web,OpenGLTextureSize + WikipediaWidth + ScrollbarWidth,P.Y);
    Height := 50;
    TControl (wbBrowser).Visible := Boolean (1);

    glEnable (GL_TEXTURE_2D);
    tOGL.Enabled := True;
  end;
end;


procedure TfrmMain.CopySubBitmapToTexture;
var
  RowLoop, ColLoop, Index : Integer;
  ScanLine : PRGBArray;
begin
  GeneratedBitmap := TBitmap.Create;
  GeneratedBitmap.PixelFormat := pf24bit;
  GeneratedBitmap.Width := OpenGLTextureSize;
  GeneratedBitmap.Height := OpenGLTextureSize;

  { SourceBitmap is our massive 512xLargeNumber bitmap. Copy part of it... }
  GeneratedBitmap.Canvas.CopyRect (Rect (0,0,OpenGLTextureSize,OpenGLTextureSize),SourceBitmap.Canvas,Rect (0,YOffset,OpenGLTextureSize,OpenGLTextureSize + YOffset));

  { Copy the TBitmap's pixels to an array of Byte. }
  Index := 0;
  for RowLoop := GeneratedBitmap.Height - 1 downto 0 do
  begin
    ScanLine := GeneratedBitmap.ScanLine[RowLoop];
    for ColLoop := 0 to GeneratedBitmap.Width - 1 do
    begin
      GeneratedByteArray[Index] := ScanLine[ColLoop].rgbtRed;
      GeneratedByteArray[Index + 1] := ScanLine[ColLoop].rgbtGreen;
      GeneratedByteArray[Index + 2] := ScanLine[ColLoop].rgbtBlue;
      Inc (Index,3);
    end;
  end;

  { Pass this Byte array to our texture... }
  glTexImage2D (GL_TEXTURE_2D,0,GL_RGB,OpenGLTextureSize,OpenGLTextureSize,0,GL_RGB,GL_UNSIGNED_BYTE,GeneratedByteArray);
  GeneratedBitmap.Free;
end;


procedure TfrmMain.FormKeyDown ( Sender : TObject;
                                 var Key : Word;
                                 Shift : TShiftState
                               );
begin
  if (Key = VK_DOWN) then
  begin
    Inc (YOffset,10);
    if (YOffset > (P.Y - (OpenGLTextureSize + HeaderHeight + FooterHeight))) then
    begin
      YOffset := (P.Y - (OpenGLTextureSize + HeaderHeight + FooterHeight));
    end;
    CopySubBitmapToTexture;
  end else
  if (Key = VK_UP) then
  begin
    Dec (YOffset,10);
    if (YOffset < 0) then
    begin
      YOffset := 0;
    end;
    CopySubBitmapToTexture;
  end else
  if (Key = VK_NEXT) then
  begin
    Inc (YOffset,OpenGLTextureSize);
    if (YOffset > (P.Y - (OpenGLTextureSize + HeaderHeight + FooterHeight))) then
    begin
      YOffset := (P.Y - (OpenGLTextureSize + HeaderHeight + FooterHeight));
    end;
    CopySubBitmapToTexture;
  end else
  if (Key = VK_PRIOR) then
  begin
    Dec (YOffset,OpenGLTextureSize);
    if (YOffset < 0) then
    begin
      YOffset := 0;
    end;
    CopySubBitmapToTexture;
  end;
end;


procedure TfrmMain.eArtistKeyDown ( Sender : TObject;
                                    var Key : Word;
                                    Shift : TShiftState
                                  );
var
  URL : String;
begin
  if (Key = VK_RETURN) then
  begin
    URL := WikiRoot + WikiPath + ANSIReplaceText (eArtist.Text,' ','+') + WikiParams;
    wbBrowser.Navigate (URL);
    { Reset the view to the top of the Wikipedia page... }
    YOffset := 0;
  end;
end;


end.
